# ubuntu_sso.gui - GUI for login and registration
#
# Author: Natalia Bidart <natalia.bidart@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""User registration GUI."""

import os
import re

import gettext
import gtk

_ = gettext.gettext


HELP_TEXT_COLOR = gtk.gdk.color_parse("#9c9a94")


class LabeledEntry(gtk.Entry):
    """An entry that displays the label within itself ina grey color."""

    def __init__(self, label, is_password=False, *args, **kwargs):
        self.label = label
        self.is_password = is_password
        self._user_input = None

        super(LabeledEntry, self).__init__(*args, **kwargs)

        self._set_label(self, None)
        self.set_tooltip_text(self.label)
        self.connect('focus-in-event', self._clear_text)
        self.connect('focus-out-event', self._set_label)
        self.show()

    def _clear_text(self, *args, **kwargs):
        """Clear text and restore text color."""
        if self._user_input is None:
            self.set_text('')
        else:
            self.set_text(self._user_input)

        self.modify_text(gtk.STATE_NORMAL, None) # restore to theme's default

        if self.is_password:
            self.set_visibility(False)

        return False # propagate the event further

    def _set_label(self, *args, **kwargs):
        """Set the proper label and proper coloring."""
        user_input = self.get_text()
        if user_input and not user_input.isspace():
            self._user_input = user_input
            return

        self._user_input = None
        self.set_text(self.label)
        self.modify_text(gtk.STATE_NORMAL, HELP_TEXT_COLOR)

        if self.is_password:
            self.set_visibility(True)

        return False # propagate the event further


class UserRegistrationGUI(object):
    """User registration GUI."""

    TC_URI = 'http://one.ubuntu.com/terms/'

    # Initial texts
    JOIN_HEADER_LABEL = _('Join') + ' <b>Ubuntu One</b>'
    NAME_ENTRY = _('Name')
    EMAIL1_ENTRY = _('Email address')
    EMAIL2_ENTRY = _('Re-type Email address')
    PASSWORD1_ENTRY = _('Password')
    PASSWORD2_ENTRY = _('Re-type Password')
    CAPTCHA_SOLUTION_ENTRY = _('Type the characters above')
    YES_TO_UPDATES = _('Yes! Email me tips and updates about Ubuntu One.')
    YES_TO_TC = _('I agree with the')
    TC = _('Terms & Conditions.')
    EMAIL_MISMATCH = _('The email addresses don\'t match, please double check '
                       'and try entering them again.')
    PASSWORD_MISMATCH = _('The passwords don\'t match, please double check '
                          'and try entering them again.')
    PASSWORD_TOO_WEAK = _('The password must have at least 8 characters, one '
                          'upper case and one number.')
    TC_NOT_ACCEPTED = _('Agreeing to the Ubuntu One Terms & Conditions is '
                        'required to subscribe')
    CAPTCHA_LOADING = _('loading captcha image...')
    VERIFY_EMAIL_LABEL = _("""Enter verification code.

A four digit verification code has just been sent t your email address.
Please enter this code below.""")
    EMAIL_TOKEN_ENTRY = _('Enter code verification here')

    def __init__(self, backend=None, close_callback=None):
        """Create the GUI and initialize widgets."""

        if backend is not None:
            self.backend = backend
            self._captcha_filename = self._mktemp()
            self.backend.generate_captcha(filename=self._captcha_filename)

        ui_filename = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                      os.pardir, 'data', 'ui.glade'))
        self.builder = gtk.Builder()
        self.builder.add_from_file(ui_filename)
        self.builder.connect_signals(self)

        self.text_widgets = ('name_entry', 'email1_entry', 'email2_entry',
                             'password1_entry', 'password2_entry',
                             'captcha_solution_entry', 'email_token_entry')
        widgets = ('join_header_label', 'join_window', 'join_cancel_button',
                   'join_ok_button', 'email_password_table', 'tc_button',
                   'yes_to_updates_checkbutton', 'yes_to_tc_checkbutton',
                   'enter_details_vbox', 'processing_vbox',
                   'captcha_loading_label', 'captcha_image', 'captcha_loading',
                   'verify_email_vbox', 'verify_email_label')
                  #+ self.text_widgets
        for name in widgets:
            widget = self.builder.get_object(name)
            assert widget is not None, '%s should be a valid object.' % name
            setattr(self, name, widget)

        for name in self.text_widgets:
            label = getattr(self, name.upper())
            is_password = name.startswith('password')
            setattr(self, name, LabeledEntry(label=label,
                                             is_password=is_password))
            assert getattr(self, name) is not None

        self.join_header_label.set_markup(self.JOIN_HEADER_LABEL)

        self.enter_details_vbox.pack_start(self.name_entry, expand=False)
        self.enter_details_vbox.reorder_child(self.name_entry, 0)
        self.email_password_table.attach(self.email1_entry, 0, 1, 0, 1)
        self.email_password_table.attach(self.email2_entry, 1, 2, 0, 1)
        self.email_password_table.attach(self.password1_entry, 0, 1, 1, 2)
        self.email_password_table.attach(self.password2_entry, 1, 2, 1, 2)
        self.enter_details_vbox.pack_start(self.captcha_solution_entry,
                                           expand=False)
        self.enter_details_vbox.reorder_child(self.captcha_solution_entry, 3)

        self._set_captcha_loading()

        self.yes_to_updates_checkbutton.set_label(self.YES_TO_UPDATES)
        self.yes_to_tc_checkbutton.set_label(self.YES_TO_TC)
        self.tc_button.set_label(self.TC)
        self.tc_button.set_uri(self.TC_URI)

        self.verify_email_label.set_text(self.VERIFY_EMAIL_LABEL)
        self.verify_email_vbox.pack_start(self.email_token_entry, expand=False)
        self.verify_email_vbox.reorder_child(self.email_token_entry, 2)

        self.warning_dialog = gtk.MessageDialog(type=gtk.MESSAGE_WARNING,
                                                buttons=gtk.BUTTONS_CLOSE,
                                                flags=gtk.DIALOG_MODAL)

        self.close_callback = close_callback
        # XXX: hack to not have the focus on the name entry
        self.join_cancel_button.grab_focus()

        self.enter_details_vbox.show()
        self.processing_vbox.hide()
        self.verify_email_vbox.hide()
        self.join_window.show()

        # hack until backend notifies the success of a registration
        #self.on_registration_successful()

    def _set_captcha_loading(self):
        """Present a spinner to the user while the captcha is downloaded."""
        self.join_ok_button.set_sensitive(False)
        spinner = gtk.Spinner()
        spinner.start()
        self.captcha_loading.pack_start(spinner)
        self.captcha_loading_label.set_text(self.CAPTCHA_LOADING)
        self.captcha_loading.show_all()

    def _set_processing_screen(self):
        """Present a spinner to the user while the request is processed."""
        spinner = gtk.Spinner()
        spinner.start()
        self.processing_vbox.pack_start(spinner)
        self.processing_vbox.show_all()

    def _mktemp(self):
        """Create a temporary filename."""
        return '/tmp/captcha.tmp'

    def debug(self, *args, **kwargs):
        """Generic function for debugging."""
        print self, args, kwargs

    def on_close_clicked(self, *args, **kwargs):
        """Call self.close_callback if defined."""
        if self.close_callback is not None:
            self.close_callback()

    def run(self):
        """Run the application."""
        gtk.main()

    def on_join_ok_button_clicked(self, *args, **kwargs):
        """Submit info for processing, present the processing vbox."""
        # check email
        email1 = self.email1_entry.get_text()
        email2 = self.email2_entry.get_text()
        if email1 != email2:
            self.warning_dialog.set_markup(self.EMAIL_MISMATCH)
            self.warning_dialog.run()
            self.warning_dialog.hide()
            return

        # check password
        password1 = self.password1_entry.get_text()
        password2 = self.password2_entry.get_text()
        if password1 != password2:
            self.warning_dialog.set_markup(self.PASSWORD_MISMATCH)
            self.warning_dialog.run()
            self.warning_dialog.hide()
            return

        if (len(password1) < 8 or
            re.search('[A-Z]', password1) is None or
            re.search('\d+', password1) is None):
            self.warning_dialog.set_markup(self.PASSWORD_TOO_WEAK)
            self.warning_dialog.run()
            self.warning_dialog.hide()
            return

        # check T&C
        if not self.yes_to_tc_checkbutton.get_active():
            self.warning_dialog.set_markup(self.TC_NOT_ACCEPTED)
            self.warning_dialog.run()
            self.warning_dialog.hide()
            return

        self.enter_details_vbox.hide()
        self.verify_email_vbox.hide()
        self._set_processing_screen()

    def on_registration_successful(self):
        """Registration can go on, user needs to verify email."""
        self.enter_details_vbox.hide()
        self.processing_vbox.hide()
        self.verify_email_vbox.show()

    def on_captcha_generated(self, captcha_id):
        """Captcha image has been generated and is available to be shown."""
        self.join_ok_button.set_sensitive(True)
        self.captcha_image.show()
        self.captcha_image.set_from_file(self._captcha_filename)
